#include <stdio.h>
#include <math.h>
#include "rate.h"

#include "lp16.h"

/* run an FIR filter characteristic on invec, put result in outvec;
	ordinarily run every pt, but if downby > 1, downsample as well
*/
runfirfilt(float *invec, float *outvec, float *filtimp, 
	int siglength, int filtlength, int downby)
{
        register float tmpsum;
        register float *sigptr, *filtptr;
        int i, j, dindex;
        int padlength, padlength2;
	static int old_filtlength = -1;
        static float *buf = NULL;
 
	if((downby < 1)||(downby > filtlength))
	{
		fprintf(stderr,"downsampling by %d is not OK\n",
			downby);
		exit(-1);
	}

	if(old_filtlength != filtlength)
	{
		if(buf != NULL)
		{
		 // printf("freeing crap\n");fflush(stdout);
			free(buf);
		}
	// printf("1 %d %d\n",siglength,filtlength);fflush(stdout);
        	buf = (float *)calloc((siglength+filtlength+1),
			sizeof(float));
        	if(buf == (float *)NULL)
        	{
                	fprintf(stderr,"Can't calloc in runfirfilt\n");
                	exit(-1);
		}
        }
 
// printf("2\n");fflush(stdout);
        padlength = filtlength/2;
        padlength2 = filtlength - padlength; /* allows for odd
						filtlengths */
        for(i=0; i<padlength; i++)
        {
                buf[i] = 0.;
        }
        for(i=0; i<siglength; i++)
        {
                buf[padlength+i] = invec[i];
        }
        for(i=0; i<padlength2; i++)
        {
                buf[i+siglength] = 0.;
        }
 
	/* Now do actual filtering */

	dindex = 0; /* index for possibly downsampled output sequence */
        for(i=0; i<siglength; i+=downby)
        {
                tmpsum = 0;
 
		sigptr = &buf[filtlength+i];
		filtptr = &filtimp[0];
                for(j=0; j<filtlength; j++)
                {
                        tmpsum += *(filtptr++) * *(sigptr--);
                }
                outvec[dindex] = tmpsum;
		dindex++;
        }
	old_filtlength = filtlength;
}

/* extract energy envelope for invector, filter and downsample it */
void envelope(float *invec, float *dwndata, int winlength, float downby)
{
	int i,j;

	//float *lpfilt = (float*) calloc(10000,sizeof(float));
	//float lpfilt[10000];

	static int lpfiltlength;
	static old_winlength = -1;
        float *buf = NULL;

	/* Use an internal buffer so that original is not modified */
	if(old_winlength != winlength)
	{
		if(buf != NULL)
		{
			free(buf);
		}
		
        	buf = (float *)calloc((winlength),
			sizeof(float));
        	if(buf == (float *)NULL)
        	{
                	fprintf(stderr,"Can't calloc in envelope\n");
                	exit(-1);
		}
        }
 

	/* Half-wave rectify the input; note that this doesn't modify
	the vector in place in the calling routine */
        for(i=0; i<winlength; i++)
        {
                /* Half-wave rectify */
                buf[i] = (invec[i] > 0 ? invec[i] : 0 );
        }

	runfirfilt(buf, dwndata, Lpfilt, winlength, Lpfiltlength,
		(int)downby);
}

/* A peak-picking routine that has been optimized for counting
syllabic maxima. */
int getpeakonsets(float *sigvec, int length)
{
	int i, j, firstpt;
	int tmp, index;
	int start, finish;
	//int lmax[100];
	int *lmax = calloc(100000,sizeof(int)); /* List of local maxima */
	int *lmin = calloc(100000,sizeof(int)); /* List of minima that precede each max */
	int *starts = calloc(100000,sizeof(int)); /* Start points, defined as point that
			occurs when value 50% up from min towards max */
	int *newstarts = calloc(100000,sizeof(int)); 
	int *peaks = calloc(100000,sizeof(int)); /* reduced list of maxima */
	int *valleys = calloc(100000,sizeof(int)); /* reduced list of minima */
	int nlmax; /* number of local maxima */
	int pmax; /* putative max index */
	int goodpeak; /* boolean for a local max being legitimate */
	float gmax; /* global maximum */
	float expfactor = EXPTHRESH ;
	float rise, startthresh, fthresh, min;
	int nstarts;
	int reduced_nstarts;
	int reduced_nlmax;

	nlmax = 0;
	nstarts = 0;
	reduced_nstarts = 0;
	reduced_nlmax = 0;
	gmax = 0.0;

	for(i=0; i<100; i++)
	{
		lmin[i] = 0;
	}
	for(i=0; i<length; i++)
	{
		if(gmax < sigvec[i]) gmax = sigvec[i];
	}
	/* Find all the local maxima */
	for(i=1; i<(length-1); i++)
	{
		if((sigvec[i] > sigvec[i-1])&&(sigvec[i] > sigvec[i+1]))
		{
			lmax[nlmax] = i;
			nlmax++;
			if(nlmax>100000)
			  printf("Too many maximum detected\n");
			//printf("%d\n",nlmax);fflush(stdout);
		}
	}

	/* Reduce by eliminating those that are not the largest within
		MINBLANKING pts on either side */
       	for(i=0; i<nlmax; i++)
	{
		goodpeak = TRUE;
		pmax = lmax[i];
		start = (pmax > ( MINBLANKING - 1) ?
				(pmax - MINBLANKING ): 0);
		finish = (pmax < (length-( MINBLANKING -1)) ? 
			(pmax + MINBLANKING ): (length -1 ));
		for(j=start; j<finish; j++)
		{
			if(sigvec[j] > sigvec[pmax]) 
			{
				goodpeak = FALSE;
				break;
			}
		}
		if(goodpeak == TRUE)
		{
			lmax[reduced_nlmax] = pmax;
			reduced_nlmax++;
		}
	}
	nlmax = reduced_nlmax;
	for(i=0; i<nlmax; i++)
	{
	/* For each local max, find the lowest local min that precedes it */

		min = 1.0e10;
		firstpt = (i==0)? 2: lmax[i-1];
		for(j=lmax[i]; j>firstpt; j--)
		{
			if((sigvec[j-1] < sigvec[j])
				&&(sigvec[j-1]<sigvec[j-2]))
			{
				if(sigvec[j-1] < min)
				{
					lmin[i] = j-1;
					min = sigvec[j-1];
				}
			}
		}
		/* use local and global thresholds for start */
		rise = sigvec[(lmax[i])] - sigvec[(lmin[i])];

		if((rise > ( REL_GLOBALTHRESH * gmax)) &&
			(rise > ( REL_LOCALTHRESH * sigvec[(lmax[i])] )) &&
			(rise > ( ABS_GLOBALTHRESH  )) 
				)
		{
			startthresh = sigvec[(lmin[i])] + .1 * rise;

			for(j=lmin[i]; j<lmax[i]; j++)
			{
				if(sigvec[j] >= startthresh)
				{
					starts[nstarts] = j;
					peaks[nstarts] = lmax[i];
					valleys[nstarts] = lmin[i];
					nstarts++;
					break;
				}
			}
		}
	}
	/* Now reduce further by putting in a blanking time related
		to an exponentially decreasing threshold */

	if(nstarts > 0)
	peaks[nstarts] = peaks[nstarts-1] + 1;

	fthresh = 0.0;
	for(i=0; i<nstarts; i++)
	{
		index = peaks[i];
		newstarts[reduced_nstarts] = starts[i];
		if(fthresh < sigvec[index])
		{
			/* Need strict blanking period, in frames */
			if((i==0)||((starts[i]-starts[i-1])> MINBLANKING))
			{
				fthresh = sigvec[index];
				reduced_nstarts++;
			}
		}
		for(j=peaks[i]; j<peaks[i+1]; j++)
		{
			fthresh *= expfactor;
		}
	}
	nstarts = reduced_nstarts;
	free(lmax);
	free(lmin);
        free(starts);
	free(newstarts);
	free(peaks);
	free(valleys);
	return(nstarts);
}

/* compute average product over pairs in a pointwise manner */
void covar_pairs(float *fvec[] , int nvecs, int length, float *outvec)
{
	int i, j, samp;
	int npairs;
	float sum, save;

	npairs = (nvecs)*(nvecs-1)/2;

	for(samp=0; samp<length; samp++)
	{
		sum = 0.;
		for(i=0; i<(nvecs-1); i++)
		{
			for(j=i+1; j<nvecs; j++)
			{
				sum += (fvec[i][samp] * fvec[j][samp]);
			}
		}
		sum /= (float)npairs;
		outvec[samp] = sum;
	}
	/* Do a simple .25 .5 .25 smoothing */

	save = outvec[0];

	for(samp=2; samp<length; samp++)
	{
		outvec[samp-2] = save;

		save = (.25 * outvec[samp])
				+ (.5 * outvec[samp-1])
				+ (.25 * outvec[samp-2]);
	}
	outvec[length-2] = save;
}
